﻿using System;
using System.Collections.Generic;

namespace PartialFraction
{
    public class QRAlgorithm
    {
        public TMatrix m;  // matrix to compute the eigenvalues from
        public TMatrix r;  // rotation matrix

        private double[] cos;  // angle for rotation matrix
        private double[] sin;

        public double[] x;
        public double[,] y;
        public double[,] oldVal;
        public double[,] oldEigenVal;
        public List<double[]> roots = new List<double[]>();
        
        public int MaxOrder;
        public int actOrder;
        public bool bShift = true;
        double phi = 0;
        public double xk_max, xk_min; // only for the graphic
        public bool[] valid; // only for the graphic

        public QRAlgorithm(int order)
        {
            MaxOrder = order;
            actOrder = MaxOrder - 1;
            m = new TMatrix(MaxOrder); // equation Matrix
            r = new TMatrix(MaxOrder); // rotation Matrix
            y = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            oldVal = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            oldEigenVal = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            cos = new double[MaxOrder];
            sin = new double[MaxOrder];
            valid = new bool[MaxOrder];
            x = new double[MaxOrder];
            xk_max = -1E30;
            xk_min = 1E30;
        }

        public void Reseize(int order)
        {
            MaxOrder = order;
            actOrder = MaxOrder - 1;
            m = new TMatrix(MaxOrder); // equation Matrix
            r = new TMatrix(MaxOrder); // rotation Matrix
            Array.Resize(ref cos, MaxOrder);
            Array.Resize(ref sin, MaxOrder);
            Array.Resize(ref valid, MaxOrder);
            Array.Resize(ref x, MaxOrder);
            xk_max = -1E30;
            xk_min = 1E30;
        }

        private void CalcRMatrixHess(int p, int q)
        {
            int i, j;
            double radius = Math.Sqrt(m.a[p, p - 1] * m.a[p, p - 1] + m.a[q, p - 1] * m.a[q, p - 1]);
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j] = 0.0;
                }
                r.a[i, i] = 1.0;
            }
            if (radius != 0.0)
            {
                if (m.a[p, p - 1] == 0.0)
                {
                    r.a[p, p] = 0.0;
                    r.a[p, q] = 1.0;
                }
                else
                {
                    r.a[p, p] = m.a[p, p - 1] / radius;
                    r.a[p, q] = -m.a[q, p - 1] / radius;
                }
                r.a[q, q] = r.a[p, p];
                r.a[q, p] = -r.a[p, q];
            }
        }

        private void CalcRMatrixQR(int p, int q)
        {
            int i, j;
            double radius;
            radius = Math.Sqrt(m.a[p, p] * m.a[p, p] + m.a[q, p] * m.a[q, p]);
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j] = 0.0;
                }
                r.a[i, i] = 1.0;
            }
            if (radius != 0.0)
            {
                if (m.a[p, p] == 0.0)
                {
                    cos[p] = 0.0;
                    sin[p] = 1.0;
                }
                else
                {
                    cos[p] = m.a[p, p] / radius;
                    sin[p] = -m.a[q, p] / radius;
                }
                r.a[p, p] = cos[p];
                r.a[p, q] = sin[p];
                r.a[q, q] = r.a[p, p];
                r.a[q, p] = -r.a[p, q];
            }
        }


        private void CalcRMatrixRQ(int p, int q)
        {
            int i, j;
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j] = 0.0;
                }
                r.a[i, i] = 1.0;
            }
            r.a[p, p] = cos[p];
            r.a[p, q] = sin[p];
            r.a[q, q] = r.a[p, p];
            r.a[q, p] = -r.a[p, q];
        }


        private void RunHessenberg(int order)
        {
            int p, q;
            for (p = 1; p < order - 1; p++)
            {
                for (q = p + 1; q < order; q++)
                {
                    if (Math.Abs(m.a[q, p - 1]) > 1E-200)
                    {
                        CalcRMatrixHess(p, q);
                        m.MultTranspMatrixLeft(r, p, q);
                        m.MultMatrixRight(r, p, q);
                    }
                }
            }
        }


        private void RunQR(int order)
        {
            int i, p, q;
            if (bShift)
            {
                if (Math.Abs(m.a[order - 1, order - 1]) > 1E-5)
                    phi = m.a[order - 1, order - 1];
                else
                    phi = 0.2;

            }
            else
                phi = 0.0;
            for (i = 0; i < order; i++)
                m.a[i, i] = m.a[i, i] - phi;
            for (p = 0; p < order - 1; p++)
            {
                q = p + 1;
                if (Math.Abs(m.a[q, p]) > 1E-200)
                {
                    CalcRMatrixQR(p, q);
                    m.MultTranspMatrixLeft(r, p, q);
                }
            }
        }


        private void RunRQ(int maxOrder)
        {
            int i, p, q;
            for (p = 0; p < maxOrder - 1; p++)
            {
                q = p + 1;
                CalcRMatrixRQ(p, q);
                m.MultMatrixRight(r, p, q);
            }
            for (i = 0; i < maxOrder; i++)
                m.a[i, i] = m.a[i, i] + phi;
        }

        private void CalcEigenvals22()
        {
            double a, b, c;
            a = 1.0;
            b = -y[0, 0] - y[1, 1];
            c = y[0, 0] * y[1, 1] - y[0, 1] * y[1, 0];

            if ((b * b) >= (4.0 * a * c))
            {
                y[0, 0] = (-b + Math.Sqrt((b * b) - (4.0 * a * c))) / 2.0 / a;
                y[1, 1] = (-b - Math.Sqrt((b * b) - (4.0 * a * c))) / 2.0 / a;
                y[0, 1] = 0.0;
                y[1, 0] = 0.0;
            }
            else
            {
                y[0, 0] = -b / 2.0 / a;
                y[1, 1] = y[0, 0];
                y[0, 1] = Math.Sqrt((4.0 * a * c) - (b * b)) / 2.0 / a;
                y[1, 0] = -y[0, 1];
            }
        }


        private bool IsDeltoOk(int order)
        {
            double delta = 10.0;
            double deltaEigen = 10.0;
            bool result = false;
            if (Math.Abs(m.a[order - 1, order - 2]) < 1E-20) // real eigenvalue
            {
                delta = m.a[order - 1, order - 1] - oldVal[0, 0];
                oldVal[0, 0] = m.a[order - 1, order - 1];
                result = Math.Abs(delta) < 1E-10;
            }
            else
            {
                // convergence of the values themselves
                y[0, 0] = m.a[order - 2, order - 2];
                y[0, 1] = m.a[order - 2, order - 1];
                y[1, 0] = m.a[order - 1, order - 2];
                y[1, 1] = m.a[order - 1, order - 1];

                delta = (y[0, 0] * y[0, 0] + y[0, 1] * y[0, 1]) -
                             (oldVal[0, 0] * oldVal[0, 0] + oldVal[0, 1] * oldVal[0, 1]) +
                             (y[1, 0] * y[1, 0] + y[1, 1] * y[1, 1]) -
                             (oldVal[1, 0] * oldVal[1, 0] + oldVal[1, 1] * oldVal[1, 1]);
                oldVal[0, 0] = y[0, 0];
                oldVal[0, 1] = y[0, 1];
                oldVal[1, 0] = y[1, 0];
                oldVal[1, 1] = y[1, 1];
                // convergence of the eigenvalues
                CalcEigenvals22();

                deltaEigen = (y[0, 0] * y[0, 0] + y[0, 1] * y[0, 1]) -
                             (oldEigenVal[0, 0] * oldEigenVal[0, 0] + oldEigenVal[0, 1] * oldEigenVal[0, 1]) +
                             (y[1, 0] * y[1, 0] + y[1, 1] * y[1, 1]) -
                             (oldEigenVal[1, 0] * oldEigenVal[1, 0] + oldEigenVal[1, 1] * oldEigenVal[1, 1]);
                oldEigenVal[0, 0] = y[0, 0];
                oldEigenVal[0, 1] = y[0, 1];
                oldEigenVal[1, 0] = y[1, 0];
                oldEigenVal[1, 1] = y[1, 1];
                if (Math.Abs(delta) > Math.Abs(deltaEigen))
                {
                    result = Math.Abs(deltaEigen) < 1E-15;
                }
                else
                {
                    result = Math.Abs(delta) < 1E-15;
                }
            }
            return result;
        }

        public int run()  // run the QR decomposation
        {
            int i;
            int count = 0;
            actOrder = MaxOrder - 1;
            double[] actRoot;
            double[] oldVal = new double[actOrder];
            double delta = 10.0;
            roots.Clear();
            RunHessenberg(actOrder);

            for (i = 0; i < actOrder; i++)
                oldVal[i] = m.a[i, i];

            while (actOrder > 1)
            {
                delta = 0;
                count++;
                RunQR(actOrder);
                RunRQ(actOrder);
                for (i = 0; i < actOrder; i++)
                {
                    delta = delta + Math.Abs(oldVal[i] - m.a[i, i]);
                    oldVal[i] = m.a[i, i];
                }
                if (IsDeltoOk(actOrder))
                {
                    if (actOrder > 1)
                    {
                        if (Math.Abs(m.a[actOrder - 1, actOrder - 2]) < 1E-20) // real eigenvalue
                        {
                            actRoot = new double[2];
                            actRoot[0] = 1;
                            actRoot[1] = -m.a[actOrder - 1, actOrder - 1];
                            roots.Add(actRoot);

                            valid[actOrder - 1] = true;
                            x[actOrder - 1] = m.a[actOrder - 1, actOrder - 1];
                            if (x[actOrder - 1] > xk_max)
                                xk_max = x[actOrder - 1];
                            if (x[actOrder - 1] < xk_min)
                                xk_min = x[actOrder - 1];
                            actOrder--;
                        }
                        else
                        {
                            if (Math.Round(y[1, 0], 4) == -Math.Round(y[0, 1], 4))
                            {
                                if (Math.Round(y[1, 0], 4) != 0) // no real eigenvalue
                                {  
                                    actRoot = new double[3];
                                    actRoot[0] = 1;
                                    actRoot[1] = -2 * y[0, 0];
                                    actRoot[2] = y[0, 0] * y[0, 0] + y[0, 1] * y[0, 1];
                                    roots.Add(actRoot);
                                }
                                else
                                {
                                    valid[actOrder - 2] = true;  // real eigenvalue
                                    x[actOrder - 2] = y[0, 0];
                                    valid[actOrder - 1] = true;
                                    x[actOrder - 1] = y[1, 1];
                                    if (x[actOrder - 1] > xk_max)
                                        xk_max = x[actOrder - 1];
                                    if (x[actOrder - 1] < xk_min)
                                        xk_min = x[actOrder - 1];

                                    if (x[actOrder - 2] > xk_max)
                                        xk_max = x[actOrder - 2];
                                    if (x[actOrder - 2] < xk_min)
                                        xk_min = x[actOrder - 2];

                                    actRoot = new double[2];
                                    actRoot[0] = 1;
                                    actRoot[1] = -y[0, 0];
                                    roots.Add(actRoot);
                               
                                    actRoot = new double[2];
                                    actRoot[0] = 1;
                                    actRoot[1] = -y[1, 1];
                                    roots.Add(actRoot);
                                }
                            }
                            else
                            {
                                actRoot = new double[3];
                                actRoot[0] = 1;
                                actRoot[1] = 2 * m.a[actOrder - 2, actOrder - 2];
                                actRoot[2] = m.a[actOrder - 2, actOrder - 2] * m.a[actOrder - 2, actOrder - 2] + m.a[actOrder - 2, actOrder - 2] * m.a[actOrder - 2, actOrder - 2];
                                roots.Add(actRoot);
                            }
                            actOrder -= 2;
                        }
                    }
                }
            }
            if (actOrder == 1)
            {
               x[actOrder - 1] = m.a[actOrder - 1, actOrder - 1];
                if (x[actOrder - 1] > xk_max)
                    xk_max = x[actOrder - 1];
                if (x[actOrder - 1] < xk_min)
                    xk_min = x[actOrder - 1];

                actRoot = new double[2];
                actRoot[0] = 1;
                actRoot[1] = -m.a[actOrder - 1, actOrder - 1];
                roots.Add(actRoot);

                valid[actOrder - 1] = true;
            }
            return count;
        }
    }
}
